@@ -93,7 +93,7 @@ data FileMetadata = FileMetaData
     , createdBy :: Maybe String
     , columnOrders :: [ColumnOrder]
     , encryptionAlgorithm :: EncryptionAlgorithm
-    , footerSigningKeyMetadata :: [Word8]
+    , footerSigningKeyMetadata :: BS.ByteString
     }
     deriving (Show, Eq)
 
@@ -149,7 +149,7 @@ defaultMetadata =
         , createdBy = Nothing
         , columnOrders = []
         , encryptionAlgorithm = ENCRYPTION_ALGORITHM_UNKNOWN
-        , footerSigningKeyMetadata = []
+        , footerSigningKeyMetadata = BS.empty
         }
 
 data ColumnMetaData = ColumnMetaData
@@ -182,7 +182,7 @@ data ColumnChunk = ColumnChunk
     , columnChunkColumnIndexOffset :: Int64
     , columnChunkColumnIndexLength :: Int32
     , cryptoMetadata :: ColumnCryptoMetadata
-    , encryptedColumnMetadata :: [Word8]
+    , encryptedColumnMetadata :: BS.ByteString
     }
     deriving (Show, Eq)
 
@@ -234,7 +234,16 @@ emptyColumnMetadata =
 
 emptyColumnChunk :: ColumnChunk
 emptyColumnChunk =
-    ColumnChunk "" 0 emptyColumnMetadata 0 0 0 0 COLUMN_CRYPTO_METADATA_UNKNOWN []
+    ColumnChunk
+        ""
+        0
+        emptyColumnMetadata
+        0
+        0
+        0
+        0
+        COLUMN_CRYPTO_METADATA_UNKNOWN
+        BS.empty
 
 emptyKeyValue :: KeyValue
 emptyKeyValue = KeyValue{key = "", value = ""}
@@ -754,13 +763,23 @@ readEncryptionAlgorithm buf pos lastFieldId = do
         Just (elemType, identifier) -> case identifier of
             1 -> do
                 readAesGcmV1
-                    (AesGcmV1{aadPrefix = [], aadFileUnique = [], supplyAadPrefix = False})
+                    ( AesGcmV1
+                        { aadPrefix = BS.empty
+                        , aadFileUnique = BS.empty
+                        , supplyAadPrefix = False
+                        }
+                    )
                     buf
                     pos
                     0
             2 -> do
                 readAesGcmCtrV1
-                    (AesGcmCtrV1{aadPrefix = [], aadFileUnique = [], supplyAadPrefix = False})
+                    ( AesGcmCtrV1
+                        { aadPrefix = BS.empty
+                        , aadFileUnique = BS.empty
+                        , supplyAadPrefix = False
+                        }
+                    )
                     buf
                     pos
                     0
@@ -792,21 +811,21 @@ readAesGcmCtrV1 v@(AesGcmCtrV1 aadPrefix aadFileUnique supplyAadPrefix) buf pos
         Just (elemType, identifier) -> case identifier of
             1 -> do
                 aadPrefix <- readByteString buf pos
-                readAesGcmCtrV1 (v{aadPrefix = aadPrefix}) buf pos lastFieldId
+                readAesGcmCtrV1 (v{aadPrefix = aadPrefix}) buf pos identifier
             2 -> do
                 aadFileUnique <- readByteString buf pos
                 readAesGcmCtrV1
                     (v{aadFileUnique = aadFileUnique})
                     buf
                     pos
-                    lastFieldId
+                    identifier
             3 -> do
                 supplyAadPrefix <- readAndAdvance pos buf
                 readAesGcmCtrV1
                     (v{supplyAadPrefix = supplyAadPrefix == compactBooleanTrue})
                     buf
                     pos
-                    lastFieldId
+                    identifier
             _ -> return ENCRYPTION_ALGORITHM_UNKNOWN
 readAesGcmCtrV1 _ _ _ _ =
     error "readAesGcmCtrV1 called with non AesGcmCtrV1"
@@ -824,17 +843,17 @@ readAesGcmV1 v@(AesGcmV1 aadPrefix aadFileUnique supplyAadPrefix) buf pos lastFi
         Just (elemType, identifier) -> case identifier of
             1 -> do
                 aadPrefix <- readByteString buf pos
-                readAesGcmV1 (v{aadPrefix = aadPrefix}) buf pos lastFieldId
+                readAesGcmV1 (v{aadPrefix = aadPrefix}) buf pos identifier
             2 -> do
                 aadFileUnique <- readByteString buf pos
-                readAesGcmV1 (v{aadFileUnique = aadFileUnique}) buf pos lastFieldId
+                readAesGcmV1 (v{aadFileUnique = aadFileUnique}) buf pos identifier
             3 -> do
                 supplyAadPrefix <- readAndAdvance pos buf
                 readAesGcmV1
                     (v{supplyAadPrefix = supplyAadPrefix == compactBooleanTrue})
                     buf
                     pos
-                    lastFieldId
+                    identifier
             _ -> return ENCRYPTION_ALGORITHM_UNKNOWN
 readAesGcmV1 _ _ _ _ =
     error "readAesGcmV1 called with non AesGcmV1"
@@ -1101,10 +1120,10 @@ readDecimalType precision scale buf pos lastFieldId = do
         Just (elemType, identifier) -> case identifier of
             1 -> do
                 scale' <- readInt32FromBuffer buf pos
-                readDecimalType precision scale' buf pos lastFieldId
+                readDecimalType precision scale' buf pos identifier
             2 -> do
                 precision' <- readInt32FromBuffer buf pos
-                readDecimalType precision' scale buf pos lastFieldId
+                readDecimalType precision' scale buf pos identifier
             _ -> error $ "UNKNOWN field ID for DecimalType" ++ show identifier
 
 readTimeType ::
@@ -1117,15 +1136,14 @@ readTimeType ::
 readTimeType isAdjustedToUTC unit buf pos lastFieldId = do
     fieldContents <- readField buf pos lastFieldId
     case fieldContents of
-        Nothing -> return (TimeType isAdjustedToUTC unit)
+        Nothing -> return (TimeType{isAdjustedToUTC = isAdjustedToUTC, unit = unit})
         Just (elemType, identifier) -> case identifier of
             1 -> do
-                -- TODO: Check for empty
-                isAdjustedToUTC' <- (== compactBooleanTrue) <$> readAndAdvance pos buf
-                readTimeType isAdjustedToUTC' unit buf pos lastFieldId
+                let isAdjustedToUTC' = elemType == toTType compactBooleanTrue
+                readTimeType isAdjustedToUTC' unit buf pos identifier
             2 -> do
                 unit' <- readUnit TIME_UNIT_UNKNOWN buf pos 0
-                readTimeType isAdjustedToUTC unit' buf pos lastFieldId
+                readTimeType isAdjustedToUTC unit' buf pos identifier
             _ -> error $ "UNKNOWN field ID for TimeType" ++ show identifier
 
 readTimestampType ::
@@ -1138,17 +1156,15 @@ readTimestampType ::
 readTimestampType isAdjustedToUTC unit buf pos lastFieldId = do
     fieldContents <- readField buf pos lastFieldId
     case fieldContents of
-        Nothing -> return (TimestampType isAdjustedToUTC unit)
+        Nothing -> return (TimestampType{isAdjustedToUTC = isAdjustedToUTC, unit = unit})
         Just (elemType, identifier) -> case identifier of
             1 -> do
-                -- TODO: Check for empty
-                isAdjustedToUTC' <- (== compactBooleanTrue) <$> readNoAdvance pos buf
-                readTimestampType False unit buf pos lastFieldId
+                let isAdjustedToUTC' = elemType == toTType compactBooleanTrue
+                readTimestampType isAdjustedToUTC' unit buf pos identifier
             2 -> do
-                _ <- readField buf pos identifier
                 unit' <- readUnit TIME_UNIT_UNKNOWN buf pos 0
-                readTimestampType isAdjustedToUTC unit' buf pos lastFieldId
-            _ -> error $ "UNKNOWN field ID for TimestampType" ++ show identifier
+                readTimestampType isAdjustedToUTC unit' buf pos identifier
+            _ -> error $ "UNKNOWN field ID for TimestampType " ++ show identifier
 
 readUnit :: TimeUnit -> BS.ByteString -> IORef Int -> Int16 -> IO TimeUnit
 readUnit unit buf pos lastFieldId = do
@@ -1157,9 +1173,12 @@ readUnit unit buf pos lastFieldId = do
         Nothing -> return unit
         Just (elemType, identifier) -> case identifier of
             1 -> do
+                _ <- readField buf pos 0
                 readUnit MILLISECONDS buf pos identifier
             2 -> do
+                _ <- readField buf pos 0
                 readUnit MICROSECONDS buf pos identifier
             3 -> do
+                _ <- readField buf pos 0
                 readUnit NANOSECONDS buf pos identifier
             n -> error $ "Unknown time unit: " ++ show n
