@@ -14,6 +14,7 @@ import DataFrame.Internal.Column (Columnable)
 import DataFrame.Internal.Expression (Expr (..))
 
 import qualified Data.Text as T
+import System.Random
 
 -- A re-implementation of the state monad.
 -- `mtl` might be too heavy a dependency just to get
@@ -61,6 +62,12 @@ renameM expr newName = deriveM newName expr
 filterWhereM :: Expr Bool -> FrameM ()
 filterWhereM p = modifyM (D.filterWhere p)
 
+sampleM :: (RandomGen g) => g -> Double -> FrameM ()
+sampleM pureGen p = modifyM (D.sample pureGen p)
+
+takeM :: Int -> FrameM ()
+takeM n = modifyM (D.take n)
+
 filterJustM :: (Columnable a) => Expr (Maybe a) -> FrameM (Expr a)
 filterJustM (Col name) = FrameM $ \df ->
     let df' = D.filterJust name df
