@@ -71,7 +71,7 @@ derive name expr df = case interpret @a df (normalize expr) of
     Left e -> throw e
     Right (TColumn value) ->
         (insertColumn name value df)
-            { derivingExpressions = M.insert name (Wrap expr) (derivingExpressions df)
+            { derivingExpressions = M.insert name (UExpr expr) (derivingExpressions df)
             }
 
 {- | O(k) Apply a function to an expression in a dataframe and
@@ -91,7 +91,7 @@ deriveWithExpr name expr df = case interpret @a df (normalize expr) of
 deriveMany :: [NamedExpr] -> DataFrame -> DataFrame
 deriveMany exprs df =
     let
-        f (name, Wrap (expr :: Expr a)) d =
+        f (name, UExpr (expr :: Expr a)) d =
             case interpret @a df expr of
                 Left e -> throw e
                 Right (TColumn value) -> insertColumn name value d
