@@ -9,6 +9,7 @@ import qualified DataFrame.Functions as F
 import qualified DataFrame.Internal.Column as DI
 
 import Data.Function
+import DataFrame.Operators
 import Test.HUnit
 
 values :: [(T.Text, DI.Column)]
@@ -36,8 +37,8 @@ foldAggregation =
             )
             ( testData
                 & D.groupBy ["test1"]
-                & D.aggregate [F.count (F.col @Int "test2") `F.as` "test2"]
-                & D.sortBy [D.Asc "test1"]
+                & D.aggregate [F.count (F.col @Int "test2") `as` "test2"]
+                & D.sortBy [D.Asc (F.col @Int "test1")]
             )
         )
 
@@ -53,8 +54,8 @@ numericAggregation =
             )
             ( testData
                 & D.groupBy ["test1"]
-                & D.aggregate [F.mean (F.col @Int "test2") `F.as` "test2"]
-                & D.sortBy [D.Asc "test1"]
+                & D.aggregate [F.mean (F.col @Int "test2") `as` "test2"]
+                & D.sortBy [D.Asc (F.col @Int "test1")]
             )
         )
 
@@ -71,9 +72,9 @@ numericAggregationOfUnaggregatedUnaryOp =
             ( testData
                 & D.groupBy ["test1"]
                 & D.aggregate
-                    [ F.mean (F.lift (fromIntegral @Int @Double) (F.col @Int "test2")) `F.as` "test2"
+                    [ F.mean (F.lift (fromIntegral @Int @Double) (F.col @Int "test2")) `as` "test2"
                     ]
-                & D.sortBy [D.Asc "test1"]
+                & D.sortBy [D.Asc (F.col @Int "test1")]
             )
         )
 
@@ -89,8 +90,8 @@ numericAggregationOfUnaggregatedBinaryOp =
             )
             ( testData
                 & D.groupBy ["test1"]
-                & D.aggregate [F.mean (F.col @Int "test2" + F.col @Int "test2") `F.as` "test2"]
-                & D.sortBy [D.Asc "test1"]
+                & D.aggregate [F.mean (F.col @Int "test2" + F.col @Int "test2") `as` "test2"]
+                & D.sortBy [D.Asc (F.col @Int "test1")]
             )
         )
 
@@ -108,9 +109,9 @@ reduceAggregationOfUnaggregatedUnaryOp =
                 & D.groupBy ["test1"]
                 & D.aggregate
                     [ F.maximum (F.lift (fromIntegral @Int @Double) (F.col @Int "test2"))
-                        `F.as` "test2"
+                        `as` "test2"
                     ]
-                & D.sortBy [D.Asc "test1"]
+                & D.sortBy [D.Asc (F.col @Int "test1")]
             )
         )
 
@@ -127,8 +128,26 @@ reduceAggregationOfUnaggregatedBinaryOp =
             ( testData
                 & D.groupBy ["test1"]
                 & D.aggregate
-                    [F.maximum (F.col @Int "test2" + F.col @Int "test2") `F.as` "test2"]
-                & D.sortBy [D.Asc "test1"]
+                    [F.maximum (F.col @Int "test2" + F.col @Int "test2") `as` "test2"]
+                & D.sortBy [D.Asc (F.col @Int "test1")]
+            )
+        )
+
+aggregationOnNoRows :: Test
+aggregationOnNoRows =
+    TestCase
+        ( assertEqual
+            "Aggregation on DataFrame with no rows"
+            ( D.fromNamedColumns
+                [ ("test1", DI.fromList ([] :: [Int]))
+                , ("sum(test2)", DI.fromList ([] :: [Int]))
+                ]
+            )
+            ( testData
+                & D.drop 12
+                & D.groupBy ["test1"]
+                & D.aggregate
+                    [F.sum (F.col @Int "test2") `as` "sum(test2)"]
             )
         )
 
@@ -148,4 +167,7 @@ tests =
     , TestLabel
         "reduceAggregationOfUnaggregatedBinaryOp"
         reduceAggregationOfUnaggregatedBinaryOp
+    , TestLabel
+        "aggregationOnNoRows"
+        aggregationOnNoRows
     ]
